/******************************************************************************
 *
 * CHECK_NRPE.C - NRPE Plugin For Nagios
 * Copyright (c) 1999-2003 Ethan Galstad (nagios@nagios.org)
 * License: GPL
 *
 * Last Modified: 2004-05-13 by Andreas Ericsson <ae@op5.se>
 *
 * Changes:
 *   Enabled backwards compatibility.
 *   Removed process_arguments() function and a bunch of variables.
 *   Made logic go in 3 steps. Backwards compatibility is now trivial.
 *
 * Usage: check_nrpe -H <host> [-p port] [-c command] [-t seconds]
 *
 * Description:
 *
 * This plugin will attempt to connect to the NRPE daemon on the specified
 * server and port, and ask it to run the specified command.
 * It will show the output of the command, and exit with the return-code
 * of the command, as passed by the NRPE daemon on the remote host.
 *
 *****************************************************************************/

#include "../common/common.h"
#include "../common/config.h"
#include "utils.h"

#define DEFAULT_NRPE_COMMAND	"_NRPE_CHECK"	/* check version of NRPE daemon */

#ifdef HAVE_SSL
SSL_METHOD *meth;
SSL_CTX *ctx;
SSL *ssl;
#endif

int use_ssl;
int sock;
int timeout = DEFAULT_SOCKET_TIMEOUT;
char *msg = NULL;

void alarm_handler(int);
void usage(void);
void show_version(void);
int nrpe_connect(int, char *, int);
int nrpe_send_query(int, char *);
int nrpe_recv_response(int);

int main(int argc, char **argv)
{
	int res;
	char *hostname = NULL, *command_name = NULL;

	int port = DEFAULT_SERVER_PORT;
	char query[MAX_INPUT_BUFFER] = "";

	extern char *optarg;
	extern int optind, opterr, optopt;

	int argindex = 0;
	int c = 1;
	int i = 1;
	int option_index = 0;

#define optchars "H:c:a:t:p:nhl"

#ifdef HAVE_GETOPT_H
#define M_GETOPT(a, b, c) getopt_long(a, b, c, long_options, &option_index)
	static struct option long_options[]={
		{"host",    required_argument, 0, 'H'},
		{"command", required_argument, 0, 'c'},
		{"args",    required_argument, 0, 'a'},
		{"no-ssl",  no_argument,       0, 'n'},
		{"timeout", required_argument, 0, 't'},
		{"port",    required_argument, 0, 'p'},
		{"help",    no_argument,       0, 'h'},
		{"license", no_argument,       0, 'l'},
		{0, 0, 0, 0}
	};
#else
#define M_GETOPT(a, b, c) getopt(a, b, c)
#endif

	/* if no options were supplied, we die here */
	if(argc < 2) usage();

	while((c = M_GETOPT(argc, argv, optchars))) {
		if(c == EOF || !c) break;
		/* process all arguments */
		switch (c) {
			case '?':
			case 'h':
				usage();
			case 'V':
				show_version();
				break;
			case 'l':
				display_license();
				break;
			case 't':
				timeout = (unsigned int)strtoul(optarg, NULL, 0);
				break;
			case 'p':
				port = (unsigned short)strtoul(optarg, NULL, 0) & 65535;
				break;
			case 'H':
				hostname = strdup(optarg);
				break;
			case 'c':
				command_name = strdup(optarg);
				break;
			case 'a':
				argindex = i;
				break;
			case 'n':
				break;
			default:
				if(!argindex) {
					printf("%c: Bad commandline argument.\n", c);
					usage();
				}
		}
	}

	/* make sure required args were supplied */
	if(hostname == NULL) {
		printf("No hostname specified. Use the '-H' switch\n");
		usage();
	}
	
	if(timeout <= 0) {
		printf("Timeout must be greater than 0\n");
		usage();
	}
	if(port <= 0 || port > 65535) {
		printf("Port must be greater than 0, and less than 65535\n");
		usage();
	}

	memset(&query, 0, sizeof(query));
	/* AE : removed non-ANSI snprintf() call */
	if(command_name && strlen(command_name))
		strncat(query, command_name, sizeof(query));
	else
		strncat(query, DEFAULT_NRPE_COMMAND, sizeof(query));
	query[sizeof(query) - 1] = '\x0';

	/* get the command args */
	if(argindex > 0) {
		for(c = argindex; c<argc; c++) {
			i = sizeof(query) - strlen(query) - 2;
			if(i <= 0) break;

			strcat(query, "!");
			strncat(query, argv[c], i);
			query[sizeof(query) - 1] = '\x0';
		}
	}

	/* generate the CRC 32 table */
	generate_crc32_table();

#ifdef HAVE_SSL
	/* initialize SSL */
	SSL_library_init();
	SSLeay_add_ssl_algorithms();
	meth = SSLv23_client_method();
	SSL_load_error_strings();
	if((ctx = SSL_CTX_new(meth)) == NULL) {
		printf("CHECK_NRPE: Error - could not create SSL context.\n");
		exit(STATE_CRITICAL);
	}
#endif

	/* initialize alarm signal handling */
	signal(SIGALRM, alarm_handler);

	/* set socket timeout */
	alarm(timeout);

	/* try ALL versions in non-ssl mode first, since it's usually
	 * a matter of 1/10's of seconds until we can discard those
	 * that are SSL-enabled when we try to connect without crypto,
	 * while SSL_read() will block if we attempt to use it on a plaintext
	 * daemon */
	use_ssl=0;
#ifdef HAVE_SSL
	for(use_ssl=0; use_ssl<2; use_ssl++)
#endif
	for(i=1; i<=NRPE_MAX_VERSION; i++) {
//		printf("use_ssl: %d, i: %d\n", use_ssl, i);
//		printf("nrp_connect(%d, %s, %d) = ", i, hostname, port);
		res = nrpe_connect(i, hostname, port);
//		printf("%d\n", res);
		if(res == -2) {
			close(sock);
			continue;
		}
//		printf("nrpe_send_query(%d, %s) = ", i, query);
		res = nrpe_send_query(i, (char *)query);
//		printf("%d\n", res);
		if(res == -2) {
			close(sock);
			continue;
		}
//		printf("nrpe_recv_response(%d) = ", i);
		res = nrpe_recv_response(i);
//		printf("%d\n", res);
		if(res < STATE_OOB) {
			puts(msg);
			exit(res);
		}
	}
	if(!strlen(msg)) msg = "Unknown error. Are we allowed to talk to the host?";
	/* this is probably an NRPE error so we add it to the output */
	printf("CHECK_NRPE: %s\n", msg);
	return(STATE_UNKNOWN);
}	/* main() */

void alarm_handler(int sig)
{
	printf("CHECK_NRPE: Socket timeout after %d seconds.\n", timeout);
	exit(STATE_CRITICAL);
}
