#include "../common/common.h"
#include "../common/config.h"
#include "utils.h"

#define DEFAULT_NRPE_COMMAND	"_NRPE_CHECK"	/* check version of NRPE daemon */

/* let high bit flag for no_ssl */
#define NO_SSL 0x80

#ifdef HAVE_SSL
extern SSL_METHOD *meth;
extern SSL_CTX *ctx;
extern SSL *ssl;
#endif

extern int sock;
extern int use_ssl;
extern int timeout;
extern char *msg;

int nrpe_connect(int version, char *hostname, int port)
{
	int result, rc;
	
	
	if(version == 1 && use_ssl) {
		/* ssl wasn't supported until version 2, so simply return -2 and
		 * let the main control loop continue to next version */
		return(-2);
	}

	/* try to connect to the host at the given port number */
	result = my_tcp_connect(hostname, port, &sock);

#ifdef HAVE_SSL
	/* do SSL handshake, but fall back to non-ssl if we fail */
	if(result == STATE_OK && use_ssl) {
		if((ssl = SSL_new(ctx)) != NULL) {
			SSL_CTX_set_cipher_list(ctx, "ADH");
			SSL_set_fd(ssl, sock);
			if((rc = SSL_connect(ssl)) != 1) {
				close(sock);
				return (-2);
			}
		}
		else {
			printf("CHECK_NRPE: Internal Error - Could not create SSL connection structure.\n");
			exit(STATE_CRITICAL);
		}

		/* bail if we had errors */
		if(result != STATE_OK) {
			SSL_CTX_free(ctx);
		}
	}
#endif
	if(result != STATE_OK) {
		printf("Unknown connection problem\n");
		exit(STATE_CRITICAL);
	}

	return(sock);
}

int nrpe_send_query(int version, char *query)
{
	packet_v1 pkt_v1;
	packet_v2 pkt_v2;
	int calculated_crc32;
	int bytes_to_send;
	int pkt_size = 0;
	void *pkt = NULL;
	int rc = 0;

	switch(version) {
		case 1:
			pkt_size = sizeof(packet_v1);
			pkt = &pkt_v1;
			break;
		case 2:
			pkt_size = sizeof(packet_v2);
			pkt = &pkt_v2;
			break;
		default:
			return (-1);
			break;
	}

	/* pkt now points to the correct packet struct type, and
	 * pkt_size is set appropriately, so we can set the required values */
	/* clear the packet buffer */
	memset(pkt, 0, pkt_size);

	/* packet version 1 */
	if(version == 1) {
		memset(&pkt_v1, 0, sizeof(pkt_v1));
		pkt_v1.packet_type = htonl(QUERY_PACKET);
		pkt_v1.packet_version = htonl(NRPE_PACKET_VERSION_1);
		pkt_v1.buffer_length = htonl(strlen(query));
		strcpy(&pkt_v1.buffer[0], query);
		pkt = &pkt_v1;
		pkt_size = sizeof(pkt_v1);
	}

	/* packet version 2 */
	/* fill the packet with semi-random data */
	randomize_buffer((char *)&pkt_v2, sizeof(pkt_v2));
	pkt_v2.packet_version = (int16_t)htons(NRPE_PACKET_VERSION_2);
	pkt_v2.packet_type = (int16_t)htons(QUERY_PACKET);
	strncpy(&pkt_v2.buffer[0], query, MAX_PACKETBUFFER_LENGTH);
	pkt_v2.buffer[MAX_PACKETBUFFER_LENGTH - 1] = '\x0';
	/* calculate the crc 32 value of the packet */
	pkt_v2.crc32_value = (u_int32_t) 0L;
	calculated_crc32 = calculate_crc32((char *)pkt, pkt_size);
	pkt_v2.crc32_value = (u_int32_t) htonl(calculated_crc32);

	/***** ENCRYPT REQUEST *****/

	bytes_to_send = pkt_size;

	/* send the packet */
#ifdef HAVE_SSL
	if(use_ssl) {
		if(version == 2) {
			rc = SSL_write(ssl, (char *)&pkt_v2, pkt_size);
		}
		else {
			/* fallthrough. should never happen, if compatibility code
			 * is added properly */
			rc = SSL_write(ssl, pkt, pkt_size);
		}
		if(rc < 0) rc = -1;
	}
	else
#endif
	switch(version) {
		case 1:
			rc = sendall(sock, (char *)&pkt_v1, &bytes_to_send);
			break;
		case 2:
			rc = sendall(sock, (char *)&pkt_v2, &bytes_to_send);
			break;
		default:
			rc = sendall(sock, (char *)&pkt, &bytes_to_send);
			break;
	}

	if(rc == -1) {
		msg = "Error sending query to host.\n";
		return STATE_UNKNOWN;
	}
	return(rc);
}

int nrpe_recv_response(unsigned short version)
{
	int pkt_size = 0;
	void *pkt = NULL;
	packet_v1 pkt_v1;
	packet_v2 pkt_v2;
	int rc;
	int bytes_to_receive;
	int packet_crc32;
	int calculated_crc32;
	int result = 0;

	switch(version) {
		case 1:
			use_ssl = 0;
			pkt = &pkt_v1;
			pkt_size = sizeof(pkt_v1);
			break;
		case 2:
			pkt = &pkt_v2;
			pkt_size = sizeof(pkt_v2);
			break;
		default:
			return (-1);
	}

	bytes_to_receive = pkt_size;

#ifdef HAVE_SSL
	if(use_ssl) {
		rc = SSL_read(ssl, pkt, pkt_size);
		printf("SSL_read(): %d, SSL_get_error(): %d\n",
			   rc, SSL_get_error(ssl, rc));
	}
	else
#endif
	rc = recvall(sock, (char *)pkt, (int *)&bytes_to_receive, timeout);

	/* socket must close BEFORE SSL_shutdown() (weird, but true) */
	close(sock);

	/* close the connection */
#ifdef HAVE_SSL
	if(use_ssl) {
		SSL_shutdown(ssl);
		SSL_free(ssl);
		SSL_CTX_free(ctx);
	}
#endif

	/* recv() error. This is OK. Just try next version */
	if(rc == -2) {
		msg = "";
		return STATE_OOB;
	}
	else if(rc == -1) {
		msg = "Error receiving data from daemon.";
		return STATE_OOB;
	}
	else if(rc == 0) {
		/* server disconnected */
		msg = "Received 0 bytes from daemon. Check remote server logs for error messages.";
		return STATE_OOB;
	}

	/* receive underflow */
	else if(bytes_to_receive < pkt_size) {
		msg = "Receive underflow. Failed to receive entire buffer.";
		return STATE_OOB;
	}

	/* packet validation */
	if(version == 1) {
		result = ntohl(pkt_v1.result_code);
		if(!strlen(pkt_v1.buffer)) {
			msg = "No output from remote host.";
		}
		msg = pkt_v1.buffer;
		return result;
	}
	else if(version == 2) {
		packet_crc32 = ntohl(pkt_v2.crc32_value);
		pkt_v2.crc32_value = 0L;
		calculated_crc32 = calculate_crc32((char *)pkt, pkt_size);
		if(packet_crc32 != calculated_crc32) {
			msg = "Response packet had invalid CRC32.";
			puts(msg);
			return STATE_OOB;
		}
		if(ntohs(pkt_v2.packet_version) != NRPE_PACKET_VERSION_2) {
			puts(msg);
			msg = "Invalid packet version received from server.";
			return STATE_OOB;
		}
		if(ntohs(pkt_v2.packet_version) != NRPE_PACKET_VERSION_2) {
			puts(msg);
			msg = "Invalid packet type received from server.";
			return STATE_OOB;
		}
		result = (int16_t)ntohs(pkt_v2.result_code);
		pkt_v2.buffer[MAX_PACKETBUFFER_LENGTH - 1] = '\x0';
		msg = pkt_v2.buffer;
	}

	return result;
}

void show_version()	{
	printf("\n");
	printf("NRPE Plugin for Nagios\n");
	printf("Copyright (c) 1999-2003 Ethan Galstad (nagios@nagios.org)\n");
	printf("Version: %s\n", PROGRAM_VERSION);
	printf("Last Modified: %s\n", MODIFICATION_DATE);
	printf("License: GPL with exemptions (-l for more info)\n");
#ifdef HAVE_SSL
	printf("SSL/TLS Available: Anonymous DH Mode, OpenSSL 0.9.6 or higher required\n");
#endif
	printf("\n");
}

void usage()
{
	printf("Usage: check_nrpe -H <host> [-p <port>] [-t <timeout>] [-c <command>] [-a <arglist...>]\n");
	printf("\n");
	printf("Options:\n");
	printf(" <host>     = The address of the host running the NRPE daemon\n");
	printf(" [port]     = The port on which the daemon is running (default=%d)\n",
		   DEFAULT_SERVER_PORT);
	printf(" [timeout]  = Number of seconds before connection times out (default=%d)\n",
		   DEFAULT_SOCKET_TIMEOUT);
	printf(" [command]  = The name of the command that the remote daemon should run\n");
	printf(" [arglist]  = Optional arguments that should be passed to the command.  Multiple\n");
	printf("              arguments should be separated by a space.  If provided, this must be\n");
	printf("              the last option supplied on the command line.\n");
	printf("\n");
	printf("Note:\n");
	printf("This plugin requires that you have the NRPE daemon running on the remote host.\n");
	printf("You must also have configured the daemon to associate a specific plugin command\n");
	printf("with the [command] option you are specifying here.  Upon receipt of the\n");
	printf("[command] argument, the NRPE daemon will run the appropriate plugin command and\n");
	printf("send the plugin output and return code back to *this* plugin.  This allows you\n");
	printf("to execute plugins on remote hosts and 'fake' the results to make Nagios think\n");
	printf("the plugin is being run locally.\n");
	printf("\n");

	exit(STATE_UNKNOWN);
}
